// C.Pizzi 2015

/* *************************************************** */
/*     version for whole analysis                      */
/* *************************************************** */


#include "missmax.h"
#include <tr1/unordered_map>

int inline extension(const string& x,  const string& y, int i, int j, int k){
/*	
 *  Find the maximum extension with k mismatches from position i of x and position j of y 
 */	
	
        int l = 0;	// length of the extension
	int count = 0;	// counter for the number of mismatches
	
	int s1 = x.length();
	int s2 = y.length();

	while(count<k+1 && i+l<s1 && j+l<s2){
		if(x[i+l]!=y[j+l])
		        count++;
		l++;
	}
	if(count == k+1) l--; // the (k+1)-th mismatch should not be counted
	return l;
}



double inline miss_avg(const string& seq1, const string& seq2, int k, bitset<100000> const * __restrict__  bitsequence, int mode, int kacs){

	double maxavg = 0;
	double maxcor = 0;
	int ind1, ind2, kcount;
	int s1 = seq1.length();
	int s2 = seq2.length();
        
	vector<int> positionlist; // positions in seq2 of the current longest match
	vector<int> newties;
	vector<int>::iterator myIntVectorIterator;
	
	std::tr1::unordered_map<char, int> symbol;
	symbol['A'] = 0;
	symbol['C'] = 1;
	symbol['G'] = 2;
	symbol['T'] = 3;
	symbol['N'] = 4;


	int tempos, tempmax, indmax; 	
     
	
	int* imax = new int[s1]; // imax[i] is the longest match between sufS1[i] and some position in s2
	
	
	/**************************************
	 * COMPUTE imax[0] naively            *
	 **************************************/

	indmax  = 0;
	imax[0] = 0;

	for(int j=0; j<s2; j++){
	
	  tempmax = extension(seq1, seq2, 0, j, k);
		
		if (tempmax > imax[0]){
			imax[0] = tempmax;
			positionlist.clear();
			positionlist.push_back(j);
		
		}
		else if (tempmax == imax[0])
			positionlist.push_back(j);	
	}

	maxavg = imax[0];
        maxcor = imax[0];
	
	/**************************************
	 * COMPUTE imax[i] from maxcor[i-1]   *
	 **************************************/
	
	bitset<100000> help1;
	bitset<100000> help2;
	bitset<100000> help3;
	
	//double mytime1 = 0.0,mytime2 = 0.0,mytime3 = 0.0;
	//double start1, start2, start3; 
	
	
	for(int i=1; i<s1-indmax+1; i++){
	
		
		if(imax[i-1] != 0){
			
			imax[i] = imax[i-1]-1;
			
			newties.clear();
			
			// check for extensions from the previous step considering all ties
			for(myIntVectorIterator = positionlist.begin(); myIntVectorIterator != positionlist.end(); myIntVectorIterator++){			
			  tempos = *myIntVectorIterator;
			  tempmax = -1;

			  if (seq1[i-1]!=seq2[tempos]){

			    if (i-1+imax[i-1]<s1 && tempos+imax[i-1]<s2)
			         tempmax = imax[i-1]-1+1+extension(seq1,seq2,i-1+imax[i-1]+1, tempos+imax[i-1]+1, 0);

			    else tempmax = imax[i-1]-1;
					
			    if(tempmax>=imax[i]){
				
			         if(tempmax != imax[i]){
				    imax[i] = tempmax;
			            newties.clear();
				 }	
				 
				 newties.push_back(tempos+1);
			    }
				
			  }
			  else{
			        if (tempos+1<s2){
				    tempmax = imax[i-1]-1;
				    if(tempmax==imax[i])
					  newties.push_back(tempos+1);
				}	
				else tempmax = 0;									
			  }			
			}		
					
			positionlist = newties;
			
		}
		else{	
		        imax[i] = 0;
		        positionlist.clear();		
		}
				
		
		help1 = (~bitsequence[symbol[seq1[i-1]]] & (bitsequence[symbol[seq1[i+imax[i-1]-1]]] >> (imax[i-1])) & (bitsequence[symbol[seq1[i+imax[i-1]]]] >> (imax[i-1]+1)));
		
		//if (imax[i]>imax[i-1]+1) 
		//	help1 &= ((bitsequence[symbol[seq[i+imax[i-1]+1]]] >> (imax[i-1]+2)) & (bitsequence[symbol[seq[i+imax[i]-1]]] >> (imax[i])));

		for(int r=imax[i-1]+1; r<imax[i];r++)
			help1 &= (bitsequence[symbol[seq1[i+r]]] >> (r+1));
		
		
		// reset the positions with the same imax[i-1] because already considered
		for(myIntVectorIterator = positionlist.begin(); myIntVectorIterator != positionlist.end(); myIntVectorIterator++){
			help1[(*myIntVectorIterator)-1]=0;
		}
		newties = positionlist;
		               
           
                for(int j=0; j<s2-imax[i]+1;j++){
           
		        // check candidates
                	if(j==0 || help1.test(j-1)){
                		               	
                		ind1 = i+(imax[i-1]);     
                		ind2 = j+(imax[i-1]);   
                			
                		tempos=0;
                		while(seq1[ind1]==seq2[ind2] && ind1<s1 && ind2<s2){               			
                			ind1++;
                			ind2++;
                			tempos++;
                			
                		}               		
                		tempos += (imax[i-1]); 		
                		
				// full verification needed
                		if (tempos>=imax[i]){ 
                			
                			ind1=i;ind2=j;kcount=0;
                			
                		       	while(ind1<i+imax[i-1] && kcount<=k){
                				if (seq1[ind1]!=seq2[ind2]) kcount++;   
                			       	ind1++;
                				ind2++;
                			}
                		       
					if(ind1 == i+imax[i-1]){	
                				
                				if(tempos>imax[i]){
                					imax[i] = tempos;
                					positionlist.clear();
                				}
                				// add to the list of longest matching positions
                				positionlist.push_back(j);	
                			}               				
                		}
                		
                	}
                }	
                
            	
            
            	
              
            	if(imax[i] == (imax[i-1])){
		        
			        help2 = ( (~bitsequence[symbol[seq1[i-1]]]) & (~bitsequence[symbol[seq1[i+imax[i-1]]]] >> (imax[i-1]+1)) & (bitsequence[symbol[seq1[i+imax[i-1]-1]]] >> (imax[i-1])));
				
				// reset the positions with the same imax[i-1] because already considered
				for(myIntVectorIterator = newties.begin(); myIntVectorIterator != newties.end(); myIntVectorIterator++){
					//cout<<"ties already considered form the previous step: "<<*myIntVectorIterator<<endl;
					help2[(*myIntVectorIterator)-1]=0;
				}
								
				for(int j=1;j<s2-imax[i]+1;j++){
				
				         if(!help1.test(j-1) && help2.test(j-1)){
					       	
						ind1=i;ind2=j;kcount=0;
                			
						while(ind1<s1 && ind2<s2 && kcount<=k){
						
							if (seq1[ind1]!=seq2[ind2]) kcount++;   
							ind1++;
							ind2++;
						}
						if(kcount>k) ind1 = ind1-1;
						ind1--;
						tempmax = ind1-i+1;
						
						if (tempmax > imax[i]){
							imax[i] = tempmax;
							positionlist.clear();
							positionlist.push_back(j);
						}
						else if (tempmax == imax[i])
							positionlist.push_back(j);
							
			
					}
				}	
				
				
		}
		
	        
		
		if(imax[i] == (imax[i-1]-1)){
				
		       	help3 = ((~bitsequence[symbol[seq1[i-1]]]) & (bitsequence[symbol[seq1[i+imax[i]]]] >> (imax[i]+1)));
				
			// reset the positions with the same imax[i-1] because already considered
			for(myIntVectorIterator = newties.begin(); myIntVectorIterator != newties.end(); myIntVectorIterator++){
			      //cout<<"ties already considered from the previous step: "<<*myIntVectorIterator<<endl;
			      help3[(*myIntVectorIterator)-1]=0;
			}
			
				
		        int maxcheck = 4*k;
			if (maxcheck>imax[i]) maxcheck = imax[i]/2;
						
		              for(int j=1;j<s2-imax[i]+1;j++){
					
				  if(!help1.test(j-1) && help3.test(j-1)){
				       
				       ind1=i;ind2=j;kcount=0;
				       for(int h=0;h<maxcheck && kcount<k+1;h++){
					 if (seq1[ind1]!=seq2[ind2]) kcount++;
					      ind1++; 
					      ind2++;
				       }
								
				       if(kcount<k+1){
				       
					    while(ind1<s1 && ind2<s2 && kcount<=k){
						
					       if (seq1[ind1]!=seq2[ind2]) kcount++;   
					            ind1++;
						    ind2++;
					       }
					       if(kcount>k) ind1 = ind1-1;
					       ind1--;
						
					       tempmax = ind1-i+1;
				        
					       if (tempmax > imax[i]){
						        imax[i] = tempmax;
							positionlist.clear();
							positionlist.push_back(j);
						}
						else if (tempmax == imax[i]){
						       	positionlist.push_back(j);
						}
					   }
					}					
				}	
		}

		
		// if mode is 1 then the full version is run. This is exact but slower as this cases are rare, but take lot of time
        	if(mode){
		  if(imax[i] == (imax[i-1]-1)){
	        
			help2 = ((~bitsequence[symbol[seq1[i-1]]]) & (~bitsequence[symbol[seq1[i+imax[i]]]] >> (imax[i]+1)));
			// reset the positions with the same imax[i-1] because already considered 
				for(myIntVectorIterator = newties.begin(); myIntVectorIterator != newties.end(); myIntVectorIterator++){
					//cout<<"ties already considered form the previous step: "<<*myIntVectorIterator<<endl;
					help2[(*myIntVectorIterator)-1]=0;
				}
				
			int maxcheck = 4*k;
			if (maxcheck>imax[i]) maxcheck = imax[i]/2;
							
			for(int j=1;j<s2-imax[i]+1;j++){
					
			       	if(!help1.test(j-1) && help2.test(j-1)){
						
						ind1=i;ind2=j;kcount=0;
						for(int h=0;h<maxcheck && kcount<k+1;h++){
							if (seq1[ind1]!=seq2[ind2]) kcount++;
							ind1++; ind2++;
						}
												
						if(kcount<k+1){
						  while(ind1<i+imax[i] && ind2<s2 && kcount<=k){
						
							if (seq1[ind1]!=seq2[ind2]) kcount++;   
							ind1++;
							ind2++;
						  }
						  if(kcount>k) ind1 = ind1-1;
						  ind1--;
						  tempmax = ind1-i+1;
						  if (tempmax == imax[i])
						      positionlist.push_back(j);
							
						}
				}
		       }	
			
	           }	
		
	        }
            	// update the summation for the average
            	maxavg += imax[i];
            	indmax = imax[i];
		if (imax[i]>maxcor) 
		  maxcor = imax[i];
            
	}
	
	for(int i=s1-indmax+1;i<s1;i++){
		imax[i] = s1-i; 
	       	maxavg += imax[i];
	}
	
	if(kacs)
	  return maxavg/(1.0*s1);
	else 
	  return maxcor;
	
}






void dist(double** __restrict__ distanceMatrix, vector<string> const& sequences, int k, int full, int kacs){
/*
 *	sequences is the vector of the input sequences
 *	k is the number of allowed mismatches
 *	kacs is 0 if it computes the maximum, 1 if it computes the average
 *      full is 0 if it computes the heuristics, 1 if it computes the exact values
 */	
	int nseq = sequences.size();
	string seq;
	bitset<100000> alphavector[nseq][5];
	char c;

	int length1, length2; 
        double value1, value2;

	// this is a mapping with array position
	std::tr1::unordered_map<char, int> symbol;
	symbol['A'] = 0;
	symbol['C'] = 1;
	symbol['G'] = 2;
	symbol['T'] = 3;
	symbol['N'] = 4; 
	
	
		
	// read the input sequences and build the boolean vectors
	for(int i = 0; i < nseq; i++){	
		
		//cout<<i<<endl;
		seq = sequences.at(i);
		length1 = seq.length();		
		
		// set up the vector of boolean 
		for(int j = 0; j < length1; j++){
			c = seq[j];
			alphavector[i][symbol[c]][j]=1;
		}			
	        
	}
	
	
	if(kacs){ // it computes kACS, which is not symmetric
	  //   #pragma omp parallel for schedule(static, 1) collapse(2)


		for(int i = 0; i < nseq; i++){
	
		        length1 = (sequences.at(i)).length();
			
			for(int j =i+1; j < nseq; j++){			 
			 
			        length2 = (sequences.at(j)).length();
			        value1 = miss_avg(sequences.at(i), sequences.at(j), k, alphavector[j], full, 1);
			        value2 = miss_avg(sequences.at(j), sequences.at(i), k, alphavector[i], full, 1);
				distanceMatrix[i][j] = (log10(length1)/(value2) - 2.0*log10(length2)/(length2+1) + log10(length2)/(value1) - 2.0*log10(length1)/(length1+1))*0.5;								
			        distanceMatrix[j][i] = distanceMatrix[i][j];
		           
		       }
		       distanceMatrix[i][i] = 0;
		}
	}
	else{ // it computes kMax, which is symmetric
	
	  for(int i = 0; i < nseq; i++){
	
		        length1 = (sequences.at(i)).length();
			
			for(int j =i+1; j < nseq; j++){			 
			 
			        length2 = (sequences.at(j)).length();
			        value1 = miss_avg(sequences.at(i), sequences.at(j), k, alphavector[j], full, 0);
				value2 = value1;
				distanceMatrix[i][j] = (log10(length1)/(value2) - 2.0*log10(length2)/(length2+1) + log10(length2)/(value1) - 2.0*log10(length1)/(length1+1))*0.5;								
			        distanceMatrix[j][i] = distanceMatrix[i][j];
		           
		       }
		       distanceMatrix[i][i] = 0;
	  }


	}	
	
		
    }
