// modified by C.Pizzi from run.cpp of kmacs by C. Leimeister

#include "parser.h"
#include "missmax.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

int main(int argc, char **argv){
	int c;
	int kacs = 1;
	int full = 0;
	while ((c = getopt (argc, argv, "fm")) != EOF)
		switch (c){
		case 'f':
		  full = 1;
		break;
		case 'm':
		  kacs = 0;
		break;
		if (isprint (optopt)){
                	fprintf (stderr, "Unknown option `-%c'.\n", optopt);
			fprintf(stderr, "usage:  missmax filename k -f -m\n");
			fprintf(stderr, " -f: run the full version \n");
			fprintf(stderr, " -m: compute maximal extension with k mismatches rather than average extension with k mismatches\n");
		}
             	else
		  fprintf (stderr,"Unknown option character `\\x%x'.\n", optopt); 
		
		
		return 1;
		default:
		abort ();
	}
    	FILE *stream;
	vector<string> taxa;
	vector<string> sequences;

        if (argc < 3) {
           fprintf(stderr, "usage: %s filename k \n", argv[0]);
           return EXIT_FAILURE;
        }
	
       	stream = fopen(argv[argc-2], "r");
	if (!stream) {
           perror("can't open input file");
	   fprintf(stderr, "usage: %s filename k len\n", argv[0]);
	   return EXIT_FAILURE;
        }
	
        int k = atoi(argv[argc-1]);
	
	double time = 0.0;
	double start = clock();

	cout << "Reading input file" << endl;
	parse(stream, taxa, sequences);
	cout << "done" << endl;

	int seqsize = sequences.size();
	double** dmat = new double*[seqsize];
		for(int i=0;i<seqsize;i++){
			dmat[i] = new double[seqsize];
	}

        dist(dmat, sequences, k, full, kacs);
        writeDmat(dmat, taxa, kacs, k);
	time += clock() - start;		
	time = time/CLOCKS_PER_SEC;
	cout << "Computation time: " << time << endl;

	if(kacs)
	  cout << "Distance matrix saved as kacs-k"<<k << endl;
	else 
	  cout << "Distance matrix saved as kmcs-k"<<k << endl;

	for(int i=0;i<seqsize;i++){
		delete [] dmat[i];
	}
	delete [] dmat;
	return EXIT_SUCCESS;
}
